
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


;;; Versions and compatibility


(defun char->integer (c &optional (integer-offset 48))
  (- (char-int c) integer-offset))


(defun digit-list->int (digit-list)
  (loop with sum = 0
        for i in digit-list do
        (setf sum (+ (* sum 10) i))
        finally (return sum)))


(defun version-string->int-list (s &optional (integer-offset 48))
  (loop with digit-list
        with digit-list-list
        with maybe-digit
        with len = (length s)
        for i below len do
        (setf maybe-digit (char->integer (aref s i) integer-offset))
        (if (and (>= maybe-digit 0)
                 (<= maybe-digit 9))
            (progn
              (push maybe-digit digit-list)
              (if (= i (1- len))
                  (push (nreverse digit-list) digit-list-list)))
            (when (or digit-list 
                      (= i (1- len)))
                  (push (nreverse digit-list) digit-list-list)
                  (setf digit-list nil)))
        finally (return (nreverse (mapcar #'digit-list->int digit-list-list)))))

       
(defun int-list>= (x y)
  (loop for tail-x on x
        for tail-y on y do
        (cond ((> (car tail-x) (car tail-y))
               (return t))
              ((= (car tail-x) (car tail-y))
               (cond ((and (null (cdr tail-x))
                           (cdr tail-y))
                      (return nil))
                     ((null (cdr tail-y))
                      (return t))
                     (t nil)))
              (t (return nil)))))


(defun cm-version>= (int-list)
  (int-list>= (version-string->int-list (cm-version-number)) int-list))


(defun cm-version= (int-list)
  (equalp (version-string->int-list (cm-version-number)) int-list))


(defmacro if-cm-version>= (int-list list-1 &optional list-2)
  (if (cm-version>= int-list)
      list-1
      list-2))


; Inclusion of functions 'pattern-last-value' and 'pattern-keyword?' 
; in cm-versions after or equal 2.4.0

; Macro 'new' in cm-versions after or equal 2.4.0 extended to
; include additional pattern classes (defined as in 2.3.4)
; and 'pval' as pattern class
 

(if-cm-version>= (2 4)
  (progn

    (defun pattern-last-value (pat)
      (pattern-value pat))

    (defun pattern-keyword? (x)
      (keyword? x))


    (defmacro new (class &body args)
      (let (type inits (expander (get class ':new-expander)))
        (if expander 
            (funcall expander args)
            (progn
              (setf type (or (find-class class) (error "No class named ~s." class))
                    inits (expand-inits type args t nil))
              `(make-instance (find-class ',class) ,@inits)))))


    (setf (get 'pval ':new-expander)
          #'(lambda (inits) 
              `(make-instance <pval> :of (lambda nil ,(car inits)))))))



(defun expand-pattern-args (inits)
  (if-cm-version>= (2 4)
      (expand-inits <pattern> inits t nil)
      (expand-pairs inits)))

