
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


;;; Sequences and Lists


(defun list-min (lst)
  (apply #'min lst))


(defun list-max (lst)
  (apply #'max lst))


(defun make-pairlist (lst)
  (loop for (x y) on lst by #'cddr 
        collect (list x y)))


(defun make-tripletlist (lst)
  (loop for (x y z) on lst by #'cdddr 
        collect (list x y z)))


(defun list->alist (lst)
  (loop with keys and data
        for (x y) on lst by #'cddr do
        (push x keys)
        (push y data)
        finally (return (pairlis keys data))))


(defun divide-listpairs (args-list avoid-keys-list &optional (pred #'eql))
 (loop with (list-1 list-2) for (x y) on args-list by #'cddr do
       (if (not (member x avoid-keys-list :test pred))
           (progn
             (push x list-1)
             (push y list-1))
           (progn
             (push x list-2)
             (push y list-2)))
       finally (return (values (nreverse list-1) (nreverse list-2)))))


(defun range-list (n &optional (start 0))
  (loop for i from start to n collect i))


(defun order-list (lst order-lst)
  (mapcar #'(lambda (x) (nth x lst)) order-lst))


(defun lifo-buffer (item lst max-length)
  (let ((len (length lst)))
   (cons item (if (>= len max-length)
                  (butlast lst (- len max-length -1))
                  lst))))


(defun buffer-horizon (avoid-item)
  (if (cl-boolean? avoid-item)
      (if avoid-item 1)
      avoid-item))


(defun doubles? (lst &optional (pred #'eql))
  (some #'(lambda (x) (member x (remove x lst :count 1 :test pred) 
                              :test pred)) lst))


(defun strong-listify (x)
  (if (or (null x)
          (not (listp x)))
      (list x)
      x))

(defun unlistify (lst)
  (if (cadr lst) lst (car lst)))


(defun ordered? (real-list)
  (do ((last-car nil (car rest))
       (rest real-list (cdr rest)))
      ((or (null rest)
           (and last-car
                (> last-car (car rest))))
       (if (null rest) t nil))))


(defun weak-ordered? (lst)
  (ordered? (remove-if (complement #'realp) lst)))


(defun ordered-real-vec? (vec &optional (plus-test nil))
  (and (arrayp vec)
       (or (not plus-test)
           (> (aref vec 0) 0))
       (loop with len = (length vec)
             for i below (1- len) do
             (if (and (= i 0)
                      (not (realp (aref vec i))))
                 (return nil)
                 (if (not (and (realp (aref vec (1+ i)))
                               (<= (aref vec i)
                                   (aref vec (1+ i)))))
                     (return nil)))
             finally (return t))))


(defun mod-position (trunc ordered-truncs round-delta)
   (do ((last-car nil (car rest))
        (rest ordered-truncs (cdr rest))
        (j 0 (incf j))) 
       ((or (null rest)
            (> (car rest) trunc))
        (cond ((null last-car)
               (cond ((<= (abs (- trunc (car rest))) round-delta)
                      (values 0 0))
                     ((<= (abs (- (+ 1 trunc) (car (last rest)))) round-delta)
                      (values (- (length ordered-truncs) 1) -1))
                     (t nil)))
              ((null rest)
               (cond ((<= (abs (- trunc last-car)) round-delta)
                      (values (- j 1) 0))
                     ((<= (abs (- trunc (+ 1 (car ordered-truncs)))) round-delta)
                      (values 0 1))
                     (t nil)))
              (t 
               (cond ((<= (abs (- trunc last-car)) round-delta)
                      (values (- j 1) 0))
                     ((<= (abs (- trunc (car rest))) round-delta)
                      (values j 0))
                     (t nil)))))))


#|   
(loop for i to 1 by 0.02 do
      (format t "~%~6F \: ~A" i 
              (multiple-value-list (mod-position i '(0.18 0.2 0.7 0.98) 0.05))))
|#



;;; Scales


; Ordered real vectors as pitch scales


(defun next-position-in-new-scale (pitch scale-step scale round-delta)
 (let (index index-offset dif
       (len (length scale)))      
  (multiple-value-setq (index index-offset)
    (loop for i from 0 below len do
        (setf dif (- (aref scale i) pitch))
        (if (> dif (- round-delta))
            (if (< dif round-delta)
                (return (values i 0))
                (if (= i 0)
                    (return (values 0 -1))
                    (if (>= (* 2 pitch) (+ (aref scale i) (aref scale (- i 1))))
                        (return (values i -1))
                        (return (values (- i 1) 1))))))
        (if (= i (- len 1))
            (return (values i 1)))))
  (setf index
        (+ index 
           scale-step
           (case index-offset
             (0 0)
             (-1 (if (> scale-step 0) -1 0))
             (1 (if (> scale-step 0) 0 1))))
        index 
        (cond ((>= index len) (- len 1))
              ((< index 0) 0)
              (t index)))
  index))


(defun next-position-in-old-scale (last-index scale-step scale)
 (let (index)      
  (setf index (+ last-index scale-step))
  (cond ((>= index (length scale)) (- (length scale) 1))
        ((< index 0) 0)
        (t index))))


;;; Numerical


(defun num-equal? (a b delta)                   
   (if (< (abs (- a b)) delta) t nil))


;;; Objects


(defun find-pattern (name)
  (find-object (write-to-string name) nil 'pattern))


(defun slot-list (class-type)
  (mapcar #'slot-definition-name (class-slots (find-class class-type))))


(defun copy-tree-obj (obj)
  (let ((clone (make-instance (type-of obj))))
     (loop for slot in (slot-list (type-of obj)) do 
           (if (slot-boundp obj slot)
               (setf (slot-value clone slot) 
                     (copy-tree (slot-value obj slot)))))
     clone))


;;; Input data check


(defun symbol-name-eql? (x y)
  (if (and (symbolp x)
           (symbolp y))
      (equalp (symbol-name x)
              (symbol-name y))))


(defun right-data? (data pred &optional n)   
  (or (funcall pred data)
      (and (consp data)
           (every pred data)
           (or (not n)
               (= n (length data))))))


(defun input-error (&rest args)
  (format t (apply #'concatenate 'string "~2%Incoherent input data: " args))
  (error ""))


(defun if-nth (n maybe-cons &optional order-list)
  (if (consp maybe-cons)
      (nth (if order-list 
               (nth n order-list)
               n)
           maybe-cons)
      maybe-cons))


(defun bnd? (x)
  (or (null x)
      (and (realp x) 
           (< x 128)
           (>= x 0))))


(defun amplitude-value? (x)
  (or (and (integerp x)
           (>= x 0)
           (<= x 127))
      (and (floatp x)
           (>= x 0)
           (<= x 1))))


(defun maybe-rhythm? (x)
  (or (and (realp x)
           (>= x 0))
      (symbolp x)))
  

;;; Misc


(defun cl-boolean? (x)
   (cond ((null x) t)
         ((eql x t) t)
         (t nil)))


(defun repl-seq-items (seq pairs &optional (test #'eql))
  (loop for i below (length seq) do
        (loop for pair-tail on pairs by #'cddr 
              while (if (funcall test (car pair-tail) (elt seq i))
                        (progn
                          (setf (elt seq i) (cadr pair-tail))
                          nil)
                        t)))
  seq)


(defun map-anyway (obj fn)   ; fn must optionally take an index
  (let ((len (if (or (listp obj)
                     (arrayp obj))
                 (length obj)
                 0)))
    (if (= len 0)
        (funcall fn obj)
        (coerce (loop for i from 0 below len
                      collect (funcall fn (elt obj i) i))
                (type-of obj)))))



(defun n-permutation? (thing n)
  (and (listp thing)
       (every #'integerp thing)
       (equalp (sort (copy-list thing) #'<=)
               (range-list (1- n)))))


(defun one-group-of-length? (thing n)
  (and (listp thing)
       (not (cdr thing))
       (n-permutation? (car thing) n)))


(defun buffer-pat-ref (x)
  (buffer-horizon (pattern-last-value x)))
       

(defun make-voice-group-list (grouping density)
  (loop with gr = (if (eql grouping t)
                      (list (range-list (1- density)))
                      grouping)
        for i below density collect
        (find-if #'(lambda (x) (member i x)) gr)))


(defun symbol-of-key (key)
  (read-from-string (symbol-name key)))


(defun symbol-keywordify (sym)
  (read-from-string (concatenate 'string ":" (symbol-name sym))))


(defun v-option? (key)
  (equalp (subseq (symbol-name key) 0 2) "v-"))



