
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


;;; Inst-midi-setup


(defclass inst-midi-setup (container)
  ((parent-setup-name :accessor parent-setup-name
                      :initform nil
                      :initarg :parent-setup-name)
   (ch-prg-list :accessor ch-prg-list
                :initform nil
                :initarg :ch-prg-list)
   (ch-pitchbend-list :accessor ch-pitchbend-list                  
                      :initform nil
                      :initarg :ch-pitchbend-list)
   (route-ch-num-list :accessor route-ch-num-list                  
                      :initform nil
                      :initarg :route-ch-num-list)
   (inst-ch-list :accessor inst-ch-list
                 :initform nil
                 :initarg :inst-ch-list)
   (inst-pitchbend-list :accessor inst-pitchbend-list
                        :initform nil
                        :initarg :inst-pitchbend-list)
   (all-pitchbends-list :accessor all-pitchbends-list
                        :initform nil
                        :initarg :all-pitchbends-list)
   (inst-pitchbend-index-list :accessor inst-pitchbend-index-list
                              :initform nil
                              :initarg :inst-pitchbend-index-list)
   (inst-key-bounds-list :accessor inst-key-bounds-list
                         :initform nil
                         :initarg :inst-key-bounds-list)
   (total-pitchbend-num :accessor total-pitchbend-num
                        :initform nil
                        :initarg :total-pitchbend-num)
   (total-key-num :accessor total-key-num
                  :initform nil
                  :initarg :total-key-num)
   (inst-num :accessor inst-num
             :initform nil
             :initarg :inst-num)
   (inst-name-list :accessor inst-name-list
                   :initform nil
                   :initarg :inst-name-list)
   (name-inst-table :accessor name-inst-table
                    :initform nil
                    :initarg :name-inst-table)
   (round-delta :accessor round-delta
                :initform nil
                :initarg :round-delta)
   (indicator-vec :accessor indicator-vec
                  :initform nil
                  :initarg :indicator-vec)
   (pc-indicator-vec :accessor pc-indicator-vec
                     :initform nil
                     :initarg :pc-indicator-vec)
   (synth-bend-range :accessor synth-bend-range
                     :initform 2
                     :initarg :synth-bend-range)
   (direct-midi-out :accessor direct-midi-out
                    :initform nil
                    :initarg :direct-midi-out)))
   

(defvar *current-inst-midi-setup*)


;;; Utilities for installing the inst-midi-setup


(defun make-indicator-vecs (all-pitchbends-list)
      (let* ((pc-num (length all-pitchbends-list)))
        (values
         (make-array (* pc-num 128) :initial-element nil)
         (make-array (* pc-num 12) :initial-element nil))))


(defun make-name-inst-table (inst-name-list)
   (loop with ht = (make-hash-table)
         for inst-name in inst-name-list 
         for i from 0 do
         (setf (gethash inst-name ht) i)
         finally (return ht)))


(defun make-inst-pitchbend-list (inst-ch-list ch-pitchbend-list)
  (mapcar #'(lambda (x) (mapcar #'(lambda (y) (nth y ch-pitchbend-list)) x)) 
          inst-ch-list))


(defun make-inst-prg-list (inst-ch-list ch-prg-list)
  (mapcar #'(lambda (x) (mapcar #'(lambda (y) (nth y ch-prg-list)) x)) 
          inst-ch-list))


(defun make-all-pitchbends-list (ch-pitchbend-list round-delta)
  (sort (remove-duplicates ch-pitchbend-list 
                           :test #'(lambda (x y) (num-equal? x y round-delta))) #'<=))


(defun make-inst-pitchbend-index-list (inst-pitchbend-list all-pitchbends-list round-delta)
  (mapcar #'(lambda (x) 
              (mapcar #'(lambda (y) 
                          (position y all-pitchbends-list 
                                    :test #'(lambda (x y) (num-equal? x y round-delta)))) 
                                x)) inst-pitchbend-list))



(defun make-inst-key-bounds-list (input-key-bounds-list inst-pitchbend-list n)
  (if (or (null input-key-bounds-list)  ; expects checked data
          (and (consp input-key-bounds-list)
               (every #'integerp input-key-bounds-list)))
      (loop with (a b) = (or input-key-bounds-list '(0 127))
            for i below n collect
            (list (+ a (car (nth i inst-pitchbend-list)))
                  (+ b (car (last (nth i inst-pitchbend-list))))))
      input-key-bounds-list))


;;; Consistency check for inst-midi-setup


(defun bounds? (bounds)
  (if (or (not (and (listp bounds)
                    (= (length bounds) 2)))
          (not (every #'realp bounds))
          (not (and (<= (car bounds) (cadr bounds))))
          (not (and (>= (car bounds) 0) 
                    (<= (cadr bounds) 128))))
      nil t))


(defun ch-prg-list? (ch-prg-list)
  (cond ((not (listp ch-prg-list))
         (error "ch-prg-list must be list"))
        ((not (every #'integerp ch-prg-list))
         (error "ch-prg-list must consist of integers"))
        ((not (every #'(lambda (x) (and (>= x 0) (<= x 127))) ch-prg-list))
         (error "ch-prgs must be within (0 127)"))
        (t t)))


(defun ch-pitchbend-list? (ch-pitchbend-list)
  (cond ((not (listp ch-pitchbend-list))
         (error "ch-pitchbend-list must be list"))
        ((not (every #'realp ch-pitchbend-list))
         (error "ch-pitchbend-list must consist of reals"))
        ((not (every #'(lambda (x) (and (>= x 0) (< x 1))) ch-pitchbend-list))
         (error "ch-pitchbends must be >= 0 and < 1"))
        (t t)))


(defun route-ch-num-list? (route-ch-num-list)
  (cond ((not (listp route-ch-num-list))
         (error "route-ch-num-list must be list"))
        ((not (every #'integerp route-ch-num-list))
         (error "route-ch-num-list must consist of integers"))
        ((not (every #'(lambda (x) (and (> x 0) (<= x 16))) route-ch-num-list))
         (error "channels per route must be > 0 and <= 16"))
        ((> (length route-ch-num-list) 8)
         (error "there must not be more than 8 routes"))
        (t t)))


(defun inst-ch-list? (inst-ch-list)
 (let ((num (loop for s in (mapcar #'length inst-ch-list) sum s)))
  (cond ((not (listp inst-ch-list))
         (error "inst-ch-list must be list"))
        ((not (every #'listp inst-ch-list))
         (error "items of inst-ch-list must be lists"))
        ((not (every #'(lambda (x) (every  #'integerp x)) inst-ch-list))
         (error "items of inst-ch-list must be lists of integers"))
        ((/= (length (remove-duplicates (reduce #'union inst-ch-list))) num)
         (warn "inst-ch-list: double-mapping on some channel "))
        (t t))))


;;; Setup and cleanup processes


(defprocess prog-setup (inst-midi-setup)
  (process for ch from 0
           for pr in (ch-prg-list inst-midi-setup)
           output 
           (new midimsg time (now)
                msg (make-program-change ch pr))))


(defmethod send-prog-setup ((setup inst-midi-setup) to)
  (events (prog-setup setup) to))


(defprocess clearbend (&optional (synth-bend-range 2))
  (process for i below 16
           output 
           (new midimsg time (now)
                msg (make-pitch-bend i 0 synth-bend-range))))


(defprocess setbend (inst-midi-setup)
  (process for ch from 0
           for pb in (ch-pitchbend-list inst-midi-setup)
           output (new midimsg time (now)
                       msg (make-pitch-bend ch pb 2))))


(defun send-equal-pan ()
  (loop for ch-list in (inst-ch-list *current-inst-midi-setup*) do
        (loop for ch in ch-list do
              (if (direct-midi-out *current-inst-midi-setup*)
                  (events (new midimsg time 0 
                               msg (make-control-change ch 10 63)) "midi.port"))))) 


(defprocess send-inst-midi-setup-data (inst-list inst-midi-setup)
  (process repeat 1 do
           (loop for inst in (remove-duplicates inst-list) do
                 (loop for ch in (nth inst (inst-ch-list inst-midi-setup))  
                       for pr = (nth ch (ch-prg-list inst-midi-setup))
                       for pb = (nth ch (ch-pitchbend-list inst-midi-setup)) do
                       (output 
                        (new midimsg time (now)
                             msg (make-program-change ch pr)))
                       (output 
                        (new midimsg time (now)
                             msg (make-pitch-bend ch pb 2)))))))


;;; Making and installing the inst-midi-setup


(defun route-ch-num-list->channelmap (route-ch-num-list)
 (let (ch-map)
  (loop with counter = 0
        for i from 0
        for chan-num in route-ch-num-list do
        (loop for chan below chan-num do
              (push (list counter (list i chan))
                    ch-map)
              (incf counter)))
  (coerce (nreverse ch-map) 'vector)))


(defun route-ch-num-list->routemap (route-ch-num-list)
  (coerce 
    (loop for i below (length route-ch-num-list)
      collect (list i i)) 'vector))
       

       
(defun make-inst-midi-setup (ch-prg-list ch-pitchbend-list inst-ch-list round-delta
                                    &key key-bounds (inst-names nil) name 
                                    (route-ch-num-list (if (cm-version>= '(2 4))
                                                           '(16 16 16 16)
                                                           '(16 16))))
  (let ((inst-midi-setup)
        (inst-prg-list))
    (inst-ch-list? inst-ch-list)
    (ch-prg-list? ch-prg-list)
    (ch-pitchbend-list? ch-pitchbend-list)
    (route-ch-num-list? route-ch-num-list)
    (or (null key-bounds)
        (right-data? key-bounds #'bounds?))
    (cond ((/= (length ch-pitchbend-list)
               (length ch-prg-list))
           (error "length of ch-prg-list and ch-pitchbend-list must be equal"))
          ((> (length ch-prg-list)
              (apply '+ route-ch-num-list))
           (error "length of ch-prg-list must be >= sum of route-ch-num-list"))
          ((some #'null (list inst-ch-list ch-prg-list ch-pitchbend-list))
           (error "input-lists must not be nil"))
          ((some #'(lambda (x) (some #'(lambda (y) (>= y (length ch-prg-list))) x)) 
                 inst-ch-list)
           (error "some instrument channel has no program defined"))
          ((and inst-names
                (/= (length inst-names)
                    (length inst-ch-list)))
           (error "inst-ch-list and inst-names-list must be of same length")))
    (setf inst-midi-setup
          (make-instance 'inst-midi-setup
            :ch-prg-list ch-prg-list
            :ch-pitchbend-list ch-pitchbend-list
            :route-ch-num-list route-ch-num-list
            :inst-ch-list inst-ch-list
            :round-delta round-delta
            :inst-name-list inst-names
            :inst-key-bounds-list key-bounds
            :name name))
    (setf (inst-pitchbend-list inst-midi-setup)
          (make-inst-pitchbend-list (inst-ch-list inst-midi-setup) (ch-pitchbend-list inst-midi-setup))
          inst-prg-list
          (make-inst-prg-list inst-ch-list inst-prg-list)) 
    (if (some #'(lambda (x) (not (ordered? x))) (inst-pitchbend-list inst-midi-setup))
        (error "pitchbends must be ordered within instrument"))
    inst-midi-setup))




(defmethod install-setup ((setup inst-midi-setup))
  (setf *current-inst-midi-setup* (copy-tree-obj setup)
        (object-name *current-inst-midi-setup*) nil 
        (parent-setup-name *current-inst-midi-setup*) (object-name setup))
  (setf (all-pitchbends-list *current-inst-midi-setup*) 
        (make-all-pitchbends-list (ch-pitchbend-list *current-inst-midi-setup*) 
                                  (round-delta *current-inst-midi-setup*)))
  (setf (total-pitchbend-num *current-inst-midi-setup*)
        (length (all-pitchbends-list *current-inst-midi-setup*))
        (inst-num *current-inst-midi-setup*)
        (length (inst-ch-list *current-inst-midi-setup*)))
  (multiple-value-bind (m n)
      (make-indicator-vecs (all-pitchbends-list *current-inst-midi-setup*))
      (setf (indicator-vec *current-inst-midi-setup*) m
            (pc-indicator-vec *current-inst-midi-setup*) n))
  (setf (total-key-num *current-inst-midi-setup*)
        (length (indicator-vec *current-inst-midi-setup*))
        (inst-pitchbend-index-list  *current-inst-midi-setup*)
        (make-inst-pitchbend-index-list (inst-pitchbend-list *current-inst-midi-setup*)
                                        (all-pitchbends-list *current-inst-midi-setup*) 
                                        (round-delta *current-inst-midi-setup*)))
  (setf (inst-key-bounds-list *current-inst-midi-setup*)
        (make-inst-key-bounds-list (inst-key-bounds-list *current-inst-midi-setup*)
                                   (inst-pitchbend-list *current-inst-midi-setup*)
                                   (inst-num *current-inst-midi-setup*)))
  (if (inst-name-list *current-inst-midi-setup*)
      (setf (name-inst-table *current-inst-midi-setup*) 
            (make-name-inst-table (inst-name-list *current-inst-midi-setup*))))
  #+(or mi_d midishare) (setf (direct-midi-out *current-inst-midi-setup*) t)
  (if-cm-version>= (2 4)
        nil           
        (progn
          (midi-set-route-map (route-ch-num-list->routemap 
                               (route-ch-num-list *current-inst-midi-setup*)))
          (midi-set-channel-map (route-ch-num-list->channelmap 
                                 (route-ch-num-list *current-inst-midi-setup*)))))  
  (if (direct-midi-out *current-inst-midi-setup*)
      (events (send-inst-midi-setup-data (range-list (1- (inst-num *current-inst-midi-setup*)))
                                         *current-inst-midi-setup*) "midi.port"))
  (describe *current-inst-midi-setup*))



;;; Functions for modifying the inst-midi-setup


(defun clear-indicator-vecs ()
   (fill (indicator-vec *current-inst-midi-setup*) nil)
   (fill (pc-indicator-vec *current-inst-midi-setup*) nil))
 

(defun change-inst-prg (inst prg)
  (loop for ch in (nth inst (inst-ch-list *current-inst-midi-setup*)) do
        (setf (nth ch (ch-prg-list *current-inst-midi-setup*)) prg))
  (if (direct-midi-out *current-inst-midi-setup*)
      (send-prog-setup *current-inst-midi-setup* "midi.port")))
  
  
(defun change-ch-prg (ch prg)
  (setf (nth ch (ch-prg-list *current-inst-midi-setup*)) prg)
  (if (direct-midi-out *current-inst-midi-setup*)
      (send-prog-setup *current-inst-midi-setup* "midi.port")))
  
                                  
(defun inc-ch-prg (ch)
   (if (= (nth ch (ch-prg-list *current-inst-midi-setup*)) 127)
       (error "cannot increment channel program 127")
       (incf (nth ch (ch-prg-list *current-inst-midi-setup*))))
   (if (direct-midi-out *current-inst-midi-setup*)
       (send-prog-setup *current-inst-midi-setup* "midi.port")))


(defun dec-ch-prg (ch)
   (if (= (nth ch (ch-prg-list *current-inst-midi-setup*)) 0)
       (error "cannot decrement channel program 0")
       (decf (nth ch (ch-prg-list *current-inst-midi-setup*))))
   (if (direct-midi-out *current-inst-midi-setup*)
       (send-prog-setup *current-inst-midi-setup* "midi.port")))


(defun inc-inst-prg (inst)                                 
   (loop for ch in (nth inst (inst-ch-list *current-inst-midi-setup*)) do
         (inc-ch-prg ch)))
  

(defun dec-inst-prg (inst)                                 
   (loop for ch in (nth inst (inst-ch-list *current-inst-midi-setup*)) do
         (dec-ch-prg ch)))


(defun change-inst-bounds (inst bounds)
  (if (not (bounds? bounds))
      (error "bounds not well defined, must be list of two reals >= 0 and =< 128"))
  (if (or (>= inst (inst-num *current-inst-midi-setup*))
          (< inst 0))
      (error "inst does not exist within current setup"))
  (setf (nth inst (inst-key-bounds-list *current-inst-midi-setup*))
        bounds))


;;; Utilities refering to the global inst-midi-setup


(defun key->index (key)
  (multiple-value-bind  
    (int comma)
    (floor key)
    (multiple-value-bind
      (pos offset)
      (mod-position comma
                    (all-pitchbends-list *current-inst-midi-setup*)
                    (round-delta  *current-inst-midi-setup*))
      (if (null pos)
          (error "key not within quantization of *current-inst-midi-setup*"))
      (+ (* (total-pitchbend-num *current-inst-midi-setup*) (+ int offset)) pos))))


(defun index->key (index)
  (multiple-value-bind (div rest)
    (floor index (total-pitchbend-num *current-inst-midi-setup*))
    (values (+ div (nth rest (all-pitchbends-list *current-inst-midi-setup*))) div)))
                     

(defun pc-index (index)
  (mod index (* 12 (total-pitchbend-num *current-inst-midi-setup*))))


(defun index-pb (index)
  (nth (mod index (total-pitchbend-num *current-inst-midi-setup*)) 
       (all-pitchbends-list *current-inst-midi-setup*)))


(defun index->ch (index inst)
   (nth (position (mod index (total-pitchbend-num *current-inst-midi-setup*))
                  (nth inst (inst-pitchbend-index-list *current-inst-midi-setup*)))
        (nth inst (inst-ch-list *current-inst-midi-setup*))))


(defun show-setup ()
  (describe *current-inst-midi-setup*))


(defun refresh-setup ()
  (install-setup (find-object (parent-setup-name *current-inst-midi-setup*))))


(defun key-in-bounds? (key inst)
  (let ((bounds (nth inst (inst-key-bounds-list *current-inst-midi-setup*))))
    (and (<= key (cadr bounds))
         (>= key (car bounds)))))



(defun transpose-key-to-bounds (key bounds)
  (let ((transposed-key key))
    (cond ((and (cadr bounds)
                (> key (cadr bounds)))
           (loop while (> transposed-key (cadr bounds)) do
                 (decf transposed-key 12))
           (if (and (car bounds)
                    (< transposed-key (car bounds)))
               (error "bounds too narrow, couldn't transpose to bounds"))
           transposed-key)
          ((and (car bounds)
                (< key (car bounds)))
           (loop while (< transposed-key (car bounds)) do
                 (incf transposed-key 12))
           (if (and (car bounds)
                    (> transposed-key (cadr bounds)))
               (error "bounds too narrow, couldn't transpose to bounds"))
           transposed-key)
          (t key))))


(defun list-step-width (index-list-1 index-list-2) 
  (- (index->key (list-max index-list-2))   ; step-width of leading voice if chord
     (index->key (list-max index-list-1))))


(defun pairs-in-list? (lst &optional (pred #'eql))
  (loop for tail on lst do
        (if (member (car tail) (cdr tail) :test pred) 
            (return-from pairs-in-list? t)))
  nil)


(defun octaves-in-index-list? (index-list)        ; octaves and not primes!
  (pairs-in-list? (mapcar #'pc-index (remove-duplicates index-list)) #'=))


(defun octave-steps? (index-list-1 index-list-2)  ; any NEW entry with octave relation to last chord 
  (some #'(lambda (y) 
            (some #'(lambda (x) (and (= (pc-index x) (pc-index y)) 
                                     (/= x y)
                                     (not (member y index-list-1))))
                  index-list-1))
        index-list-2))


(defun inst-control-change-msgs (inst controller value &optional (ahead 0))
  (loop for ch in (nth inst (inst-ch-list *current-inst-midi-setup*)) 
      collect (new midimsg time (+ (now) ahead) 
                   msg (make-control-change ch controller value))))


;;; Event-class depending on *current-inst-midi-setup* 
;;; just for internal use


(defclass setup-tied-midi (event)
  ((keynum :initform nil :accessor midi-keynum
           :initarg keynum :initarg :keynum)
   (index :initform nil :accessor midi-index
          :initarg keynum :initarg :index)
   (duration :initform 0.5 :accessor midi-duration
             :initarg duration :initarg :duration)
   (by-mf-factor :initform nil :accessor by-mf-factor
                 :initarg by-mf-factor :initarg :by-mf-factor)
   (amplitude :initform 64 :accessor midi-amplitude
              :initarg amplitude :initarg :amplitude)
   (amp-array :initform nil :accessor amp-array
              :initarg amp-array :initarg :amp-array)
   (inst :initform nil :accessor midi-inst
         :initarg channel :initarg :inst)
   (setup :initform nil :accessor setup
          :initarg setup :initarg :setup)
   (voice :initform nil :accessor voice
          :initarg voice :initarg :voice)
   (print-midi :initform nil :accessor print-midi
               :initarg print-midi :initarg :print-midi)
   (output-type :initform nil :accessor output-type
                :initarg output-type :initarg :output-type)))


; Methods for class 'setup-tied-midi'


(defmethod object->midi ((object setup-tied-midi))
  (multiple-value-bind 
    (abs-key setup-tied-key)
    (index->key (midi-index object))
    (values                   
     (new midi   ; midi-object with integer keynum respecting setup-pitchbend
          time (object-time object)
          keynum setup-tied-key
          duration (midi-duration object)
          amplitude (midi-amplitude object)
          channel (index->ch (midi-index object) 
                             (midi-inst object)))
     abs-key)))



(defmethod write-event ((object setup-tied-midi) io time)
  (let (midi-obj
        abs-key
        amp)
    (when (or (print-midi object)
              (eql (output-type object) :midi))
          (multiple-value-setq (midi-obj abs-key)
            (object->midi object))
          (setf amp (or (amp-array object)
                        (midi-amplitude midi-obj))))
    (if (print-midi object)  ; print sounding key, but store pb-relative midi-key 
        (format (print-midi object) 
                (concatenate 'string "~% voice-nr~3,,,'_@A / inst-nr~3,,,'_@A:  time~11,5,,,'_F  "
                             "dur~11,5,,,'_F   key~8,,,'_@A   amp"
                             (typecase amp (float "~14,6,,,,'_F")
                                           (integer "~14,,,'_@A")
                                           (array "__~12,,,'_@A"))
                             "   ch~3,,,'_@A")
                (voice object)
                (midi-inst object)
                (* (by-mf-factor object) (object-time midi-obj))
                (* (by-mf-factor object) (midi-duration midi-obj))
                abs-key
                amp
                (midi-channel midi-obj)))
    (cond ((eql (output-type object) :midi)
           (write-event midi-obj io time))
          ((eql (output-type object) :setup-tied-midi)
           (call-next-method object io time))
          ((null (output-type object)))
          (t (error "wrong output-type")))))


;;; Predefined inst-midi-setups


(make-inst-midi-setup '(73 68 71 70 60 56 57 24 46 1 0 5 40 41 42 43)
                 '(0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0)
                  (loop for i to 15 collect (list i))
                  0.00001
                  :inst-names '(:fl :ob :cl :bs :hr :tr :tb :gt :hp :pc :pf :epf :vl :va :vc :db)
                  :name 'temp)


(make-inst-midi-setup '(73 73 68 68 71 71 0 0 70 0 40 40 41 41 43 43)
                 '(0 0.5 0 0.5 0 0.5 0 0.5 0 0 0 0.5 0 0.5 0 0.5)
                 '((0 1) (2 3) (4 5) (6 7) (8) (9) (10 11) (12 13) (14 15))
                  0.00001
                 :inst-names '(:fl :ob :cl :pf :bs :pc :vl :vc :db)
                 :name 'qt)


(make-inst-midi-setup '(73 73 73 71 71 71 70 70 70 0 40 40 40 43 43 43)
                 '(0 0.33 0.66 0 0.33 0.66 0 0.33 0.66 0 0 0.33 0.66 0 0.33 0.66)
                 '((0 1 2) (3 4 5) (6 7 8) (9) (10 11 12) (13 14 15))
                  0.07
                 :inst-names '(:fl :cl :bs :pc :vl :db)
                 :name 'st)

; (install-setup (find-object "temp"))


