
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


(defvar *pe-option-conditions* (make-hash-table))


; Options that may be defined with patterns


(setf (gethash :pitch *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x)
                               (>= x 0)
                               (<= x 127)))
            "pitch must be real >= 0 and <= 127"))


(setf (gethash :step *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x)
                               (>= x -127)
                               (<= x 127)))
            "step must be real >= -127 and <= 127"))


(setf (gethash :rhythm *pe-option-conditions*)
      (cons #'(lambda (x) (or (maybe-rhythm? x)
                              (and (arrayp x)
                                   (= (length x) 1)
                                   (maybe-rhythm? (aref x 0)))))
            "rhythm must be (at least) symbol or real >= 0 or array of one element "))


(setf (gethash :dyn *pe-option-conditions*)
      (cons #'(lambda (x) (or (amplitude-value? x)
                              (nth-value 3 (parse-dyn-array x :with-error nil))))
            "dyn must be integer >= 0 and <= 127 or float >= 0 and <= 1 or dynamic array of correct syntax"))


(setf (gethash :max-next *pe-option-conditions*)
      (cons #'(lambda (x) (and (integerp x) (> x 1)))
            "max-next must be integer > 1"))


(setf (gethash :lb *pe-option-conditions*)
      (cons #'(lambda (x) (bnd? x))
            "lb must be integer between 0 and 127 or nil"))


(setf (gethash :ub *pe-option-conditions*)
      (cons #'(lambda (x) (bnd? x))
            "ub must be integer between 0 and 127 or nil"))


(setf (gethash :transpose-to-bounds *pe-option-conditions*)
      (cons #'cl-boolean? 
            "transpose-to-bounds must be t / nil"))


(setf (gethash :max-next *pe-option-conditions*)
      (cons #'(lambda (x) (and (integerp x) (plusp x)))
            "max-next must be positive integer"))


(setf (gethash :unlock-backward-shift *pe-option-conditions*)
      (cons #'realp 
            "unlock-backward-shift must be real"))


(setf (gethash :dur-factor *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x)
                               (>= x 0)
                               (<= x 1))) 
            "dur-factor must be real >= 0 and <= 1"))


(setf (gethash :no-8-steps *pe-option-conditions*)
      (cons #'(lambda (y) (or (cl-boolean? y)
                              (and (integerp y)
                                   (not (minusp y)))))
            "no-8-steps must be t / nil / non-negative integer"))


(setf (gethash :step-start *pe-option-conditions*)
      (cons #'(lambda (x) (flet  ((pred (y) (or (null y)
                                                (and (realp y)
                                                     (>= y 0)
                                                     (<= y 127)))))
                            (or (pred x)
                                (and (consp x)
                                     (every #'pred x)))))
            "step-start must be real >= 0 and <= 127 or list (chord) of"))


(setf (gethash :pitch-scale *pe-option-conditions*)
      (cons #'(lambda (x) (ordered-real-vec? x t)) 
            "pitch-scale must be ordered vector of positive reals"))


; Different key for runtime error checking

(setf (gethash :octave-step-horizon *pe-option-conditions*)
      (cons #'(lambda (y) (or (cl-boolean? y)
                              (and (integerp y)
                                   (not (minusp y)))))
            "no-8-steps must be t / nil / non-negative integer"))


(setf (gethash :no-reps *pe-option-conditions*)
      (cons #'(lambda (y) (or (cl-boolean? y)
                              (and (integerp y)
                                   (not (minusp y)))))
            "no-reps must be t / nil / non-negative integer"))


(setf (gethash :pitch-rep-horizon *pe-option-conditions*)
      (cons #'(lambda (y) (or (cl-boolean? y)
                              (and (integerp y)
                                   (not (minusp y)))))
            "no-reps must be t / nil / non-negative integer"))


(setf (gethash :max-step *pe-option-conditions*)
      (cons #'(lambda (x) (or (null x)
                                   (and (realp x) (plusp x))))
            "max-step must be positive real / nil"))


(setf (gethash :base *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "base must be real > 0"))


(setf (gethash :level-change-step-time *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "level-change-step-time must be real > 0"))


(setf (gethash :min-level-change-steps *pe-option-conditions*)
      (cons #'(lambda (x) (and (integerp x) (plusp x)))
            "min-level-change-steps must be integer > 0"))


(setf (gethash :level-change-delay *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "level-change-delay must be real > 0"))


(setf (gethash :level-change-break *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "level-change-break must be real > 0"))


(setf (gethash :level-shape-shrink *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x) (< x 0.95)))
            "level-shape-shrink must be real > 0 and <= 0.95"))


; Non-dynamic options


(setf (gethash :tempo *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "tempo must be real > 0"))

(setf (gethash :beat *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "beat must be real > 0"))


(setf (gethash :max-time *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "max-time must be real > 0"))


(setf (gethash :abs-times *pe-option-conditions*)
      (cons #'cl-boolean? 
            "abs-times must be t /nil"))


(setf (gethash :print *pe-option-conditions*)
      (cons #'(lambda (x) (or (cl-boolean? x) (stringp x)))
            "print must be t / nil / string"))


(setf (gethash :v-tempo *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "v-tempo must be real > 0"))


(setf (gethash :v-beat *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "v-beat must be real > 0"))


(setf (gethash :v-max-time *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (plusp x)))
            "v-max-time must be real > 0"))


(setf (gethash :v-abs-times *pe-option-conditions*)
      (cons #'cl-boolean? 
            "v-abs-times must be t /nil"))


(setf (gethash :v-print *pe-option-conditions*)
      (cons #'(lambda (x) (or (cl-boolean? x) (stringp x)))
            "v-print must be t / nil / string"))


(setf (gethash :level *pe-option-conditions*)
      (cons #'amplitude-value?
            "level must be integer >= 0 and <= 127 or float >= 0 and <= 1"))


(setf (gethash :voice-shift *pe-option-conditions*)
      (cons #'(lambda (x) (and (realp x) (>= x 0)))
            "voice-shift must be real >= 0"))


(setf (gethash :strict-step-start *pe-option-conditions*)
      (cons #'cl-boolean? 
            "strict-step-start must be t / nil"))


(setf (gethash :max-buffer-length *pe-option-conditions*)
      (cons #'(lambda (x) (and (integerp x) (plusp x))) 
            "max-buffer-length must be positive integer"))


(setf (gethash :runtime-data-check *pe-option-conditions*)
      (cons #'cl-boolean? 
            "runtime-data-check must be t / nil"))




(defun option-check (option data v-data-pats density dyn-option?) 
  (when (and (consp data)
             (/= (length data) density))
        (format t "~2%Incoherent input data, wrong list length, option: ~A~2%" option)
        (error ""))
  (when (and (pattern? data)
             (/= density 1))
        (format t (concatenate 'string "~2%Incoherent input data, option: ~A, pattern for" 
                               " all voices must be defined per virtual voice~2%") option)
        (error ""))
  (loop for datum in (listify data) do
        (when (not (or (and dyn-option?
                            (or (pattern? datum)
                                (and (symbolp datum)
                                     (not (cl-boolean? datum))
                                     (symbol-pat-reference datum v-data-pats nil))))
                       (funcall (car (gethash option *pe-option-conditions*)) datum)))
              (format t (concatenate 'string "~2%Incoherent input data, item ~A with option ~A ~2%"
                                     "Single condition " 
                                     (if dyn-option? " (if no pattern or pattern reference)" "")
                                     " : ~%~A") 
                      datum option (cdr (gethash option *pe-option-conditions*)))
              (if dyn-option?
                  (format t (concatenate 'string "~2%This option may be set to an allowed single item for all voices, ~%"
                                "a quoted symbol as v-voice pattern reference for all voices or ~%"
                                "a voice specifying list, which may contain allowed items, ~%"
                                "patterns for options to be changed per voice event ~%and quoted symbols, "
                                "referencing v-voice patterns~2%"))
                  (format t (concatenate 'string "~2%This option may be set to an allowed single item for all voices or "
                                "a voice specifying list, containing allowed items~2%")))
              (error ""))))



(defun grouping-option-error? (key val v-data-pats density)
 (let (merge)
  (when (and val
             (not (or (cl-boolean? val)
                      (and (symbolp val)
                           (not (cl-boolean? val))
                           (symbol-pat-reference val v-data-pats nil))
                      (and (listp val)
                           (every (lambda (x) (and (listp x)
                                                   (every #'integerp x))) val)
                           (and (not (doubles? (setf merge (apply #'append val))))
                                (subsetp merge (range-list (1- density))))))))
        (input-error "~2%" (symbol-name key)
                     " must be t / nil or list of lists with (not necessarily all) ~%"
                     "integers between 0 and (number-of-voices - 1), each occuring not more ~%" 
                     "than once or a (symbol) reference to a v-voice pattern, that produces such data. ~%"
                     "The partition indicates the groups of voices, within intervals should be avoided. ~%"
                     "I.e.: :no-1 '((0) (1 2)) means that primes are allowed between voices 0 and 1, ~%"
                     "between 0 and 2, but not between 1 and 2. If voice 0 gives out a chord, it ~%"
                     "is not allowed to contain primes. If there is another voice, ~%"
                     "it may be a prime to 0, 1 or 2. ~%The same with :no-8.~2%"))))



(defun pat-events-input-check (pitch-pats rhythm-pats instruments &rest args &key dyn 
                                no-8 no-1 voice-order tempo beat abs-times print 
                                transpose-to-bounds voice-shift max-time 
                                unlock-backward-shift dur-factor no-reps 
                                no-8-steps to max-next max-buffer-length
                                max-step step-start strict-step-start pitch-scale lb ub sleep
                                level base level-change-step-time min-level-change-steps 
                                level-change-delay level-change-break level-shape-shrink
                                v-data-pats v-rhythm-pats v-print 
                                v-tempo v-beat v-abs-times v-max-time v-voice-order 
                                input-data-check runtime-data-check mf-tempo mf-measure)
  (declare (ignore to pitch-scale input-data-check mf-tempo mf-measure))
  (let ((density (if (listp pitch-pats) (length pitch-pats) 1))
        (v-density (if (listp v-data-pats) (length v-data-pats) 1)))
    (loop for (key val) on 
          (list :dyn dyn :lb lb :ub ub :transpose-to-bounds transpose-to-bounds :max-next max-next
                :unlock-backward-shift unlock-backward-shift :base base :no-reps no-reps :no-8-steps no-8-steps
                :level-change-step-time level-change-step-time :min-level-change-steps min-level-change-steps
                :level-change-delay level-change-delay :level-change-break level-change-break :step-start step-start
                :level-shape-shrink level-shape-shrink :max-step max-step :dur-factor dur-factor) 
          by #'cddr do
          (when (member key args)  ; standard check for options eventually defined with patterns
              (option-check key val v-data-pats (if (v-option? key) v-density density) t)))

    (loop for (key val) on 
          (list :tempo tempo :beat beat :abs-times abs-times :max-time max-time :print print
                :v-tempo v-tempo :v-beat v-beat :v-abs-times v-abs-times :v-max-time v-max-time 
                :v-print v-print :level level :voice-shift voice-shift :strict-step-start strict-step-start
                :max-buffer-length max-buffer-length :runtime-data-check runtime-data-check)
          by #'cddr do
          (if (member key args)  ; standard check for options defined without patterns
              (option-check key val v-data-pats (if (v-option? key) v-density density) nil)))

    ; various additional checks
    
    (cond ((let ((data (list instruments pitch-pats rhythm-pats)))
             (and (not (every #'null data))
                  (or (some #'(lambda (x) (not (or (right-data? x #'pattern? density)))) (cdr data))
                      (not (apply #'= (mapcar #'(lambda (x) (length (listify x))) (cdr data)))))))
           (input-error "pitch-pats and rhythm-pats must be pattern ~%"
                        "or lists of patterns of equal length~2%"))
          
          ((and (not (every #'null (list pitch-pats rhythm-pats instruments)))
                (not (right-data? instruments #'(lambda (x) (and (integerp x)
                                                                 (>= x 0)
                                                                 (< x (inst-num *current-inst-midi-setup*))))
                                      density))) 
           (input-error "instruments must be interger or integer list right length, ~%"
                        "representing existing instruments in the current setup~2%"))
          
          ((and voice-order
                (not (n-permutation? voice-order density)))
           (input-error "voice-order must be permutation list of integers from 0 to (number-of-voices - 1)~2%"))
          ((and v-voice-order
                (not (n-permutation? v-voice-order v-density)))
           (input-error "v-voice-order must be permutation list of integers from 0 to (number-of-v-voices - 1)~2%"))

          ((and (or (member :lb args)
                    (member :ub args))
                (loop for i from 0 below density do
                      (if (not (weak-ordered? (list (if-nth i lb)
                                                    (if-nth i step-start)
                                                    (if-nth i ub))))
                          (return t))))
           (input-error "lb <= step-start <= ub must hold for all voices~2%"))

          ((and sleep
                (or (not (numberp sleep))
                    (< sleep 0)))
           (input-error "sleep must be non-negative number~2%"))

          ((and no-1 (grouping-option-error? :no-1 no-1 v-data-pats density)))
          ((and no-8 (grouping-option-error? :no-8 no-8 v-data-pats density)))
         
          ((and (member :level args)
                (listp level)
                (some #'(lambda (x) 
                          (some #'(lambda (y) 
                                    (and (intersection
                                          (nth (if-nth x instruments) 
                                               (inst-ch-list *current-inst-midi-setup*))
                                          (nth (if-nth y instruments) 
                                               (inst-ch-list *current-inst-midi-setup*)))
                                         (not (num-equal? (nth x level)
                                                          (nth y level) 0.001))))
                                (range-list (1- density) x)))
                      (range-list (1- density))))
           (input-error "attempt to set the same channel to different levels~2%"))

         
          ((and (or v-data-pats v-rhythm-pats)
                (some #'(lambda (x) (not (or (right-data? x #'pattern? v-density)))) 
                      (list v-data-pats v-rhythm-pats)))
           (input-error "v-data-pats and v-rhythm-pats must be pattern ~%"
                        "or lists of patterns of equal length~2%"))
          
          (t t))))


; Runtime data check


(proclaim '(inline runtime-item-check next-with-check))

(defun runtime-item-check (item key voice-number pat-type)
  ; pat-type 0: voice, type 1: v-voice
  (if (funcall (car (gethash key *pe-option-conditions*)) item)
      item
      (progn 
        (format t (concatenate 'string "~2%Runtime error, value ~A with key ~A in "
                               (case pat-type (0 "voice") (1 "v-voice")) 
                               " number ~A~2%Condition : ~%~A~2%") 
                item key voice-number (cdr (gethash key *pe-option-conditions*)))
        (error ""))))



(defun next-with-check (pat key voice-number pat-type)
 (let ((next-item (next pat)))
  (if (listp key)
      (loop for k in key do
            (runtime-item-check next-item k voice-number pat-type))
      (runtime-item-check next-item key voice-number pat-type))))



(defun collect-v-data-option-keys (vo-input-data v-data-pats)
  ; collects list of key-lists, the keys for each v-data-pat's runtime checks 
  (with-slots (option-key-vec option-val-vec option-num multiple-option?-vec 
               option-pat-type-vec current-value-function-vec density) 
              vo-input-data
    (loop with runtime-check = (aref option-val-vec
                                     (position :runtime-data-check option-key-vec))
          with v-data-pat-list = (listify v-data-pats)
          with option-pats = (make-list (length v-data-pat-list) :initial-element nil)
          for i below option-num do
          (if (arrayp (aref option-pat-type-vec i))
              (loop for j below density do
                    (if (and (or (eql runtime-check t)
                                 (and (arrayp runtime-check)
                                      (aref runtime-check j)))
                             (eql 1 (aref (aref option-pat-type-vec i) j)))
                        (pushnew (aref option-key-vec i)
                                 (nth (position (aref (aref option-val-vec i) j) v-data-pat-list)
                                      option-pats))))
              (when (and runtime-check
                         (eql 1 (aref option-pat-type-vec i)))
                    (push (aref option-key-vec i)
                          (nth (position (aref option-val-vec i) v-data-pat-list)
                               option-pats))))
          finally (return option-pats))))
