
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


;;; Utility processes 


(defprocess lock-index (index voice-number)
   (process repeat 1 do 
            (push voice-number (aref (indicator-vec *current-inst-midi-setup*) index))))

 
(defprocess unlock-index (index voice-number)
   (process repeat 1 do 
            (setf (aref (indicator-vec *current-inst-midi-setup*) index)
                  (remove voice-number (aref (indicator-vec *current-inst-midi-setup*) index)
                          :count 1))))        ; there could be primes of one chord 


(defprocess lock-pc-index (index voice-number)
   (process repeat 1 do 
            (push voice-number 
                  (aref (pc-indicator-vec *current-inst-midi-setup*) (pc-index index)))))


(defprocess unlock-pc-index (index voice-number)
   (process repeat 1 do 
            (setf (aref (pc-indicator-vec *current-inst-midi-setup*) (pc-index index))
                  (remove voice-number (aref (pc-indicator-vec *current-inst-midi-setup*) (pc-index index))
                          :count 1))))       ; there could be primes or octaves of one chord 


(defprocess mix-processes (process-list shift-list)   
  (process repeat 1 do
           (mapcar #'(lambda (x y) (sprout x (+ (now) y))) process-list shift-list)))


;;; Looking up pitch status


(defun primes-in-group? (index avoid-primes-group)                                        
   (some #'(lambda (x) (member x avoid-primes-group))
         (aref (indicator-vec *current-inst-midi-setup*) index)))


(defun real-octaves-in-group? (index avoid-octaves-group)                                        
   (some #'(lambda (x) (member x avoid-octaves-group))
         (set-difference (aref (pc-indicator-vec *current-inst-midi-setup*) (pc-index index))
                         (aref (indicator-vec *current-inst-midi-setup*) index))))


;;; Time parsing


(defun time-value (x)                         
  (if (arrayp x) (aref x 0) x))   ; rests marked as vectors


(defun time-parse (x &optional (tempo 60) (beat 0.25) abs-times)
  (typecase x (number (if abs-times
                          (coerce x 'float)
                          (rhythm x tempo beat)))
              (t (rhythm x tempo beat))))



;;; Utilities for single voice process


(defun nth-option-val-vec (option-val-vec multiple-option?-vec i)
  (map 'array #'(lambda (x y) (if y (aref x i) x)) 
       option-val-vec multiple-option?-vec))


(defun nth-current-value-function-vec (current-value-function-vec i)
  (let ((a (make-array *vo-option-num* :element-type 'function)))
    (loop for k below *vo-option-num* do
          (setf (aref a k) 
                (vo-current-value-function-ref current-value-function-vec k i)))
    a))

(defun nth-event-option-key+pat-alist (vo-input-data i)   
  ; gives alist of keys and event option patterns in input-order for voice number i
 (with-slots (option-key-vec input-order->vo-key-pos-list) vo-input-data
  (loop for pos in input-order->vo-key-pos-list 
        if (and pos (funcall #'(lambda (x) (and (numberp x) (= x 0)))
                             (vo-option-pat-type-ref vo-input-data pos i)))
        collect (cons (aref option-key-vec pos) (vo-option-val-ref vo-input-data pos i)))))



;;; Macro for specified access to options given by 'pat-events'  (values / patterns)


(defmacro with-dynamic-vo-symbol-reference (&rest vo-processform)
  `(with-slots (option-key-vec option-val-vec current-value-function-vec 
                input-order->vo-key-pos-list
                option-pat-type-vec multiple-option?-vec density option-num) 
               vo-input-data
    (let* ((voice-option-val-vec 
            (nth-option-val-vec option-val-vec multiple-option?-vec voice-number))
           (voice-event-option-key+pat-alist
            (nth-event-option-key+pat-alist vo-input-data voice-number))
           voice-current-value-function-vec)
      (declare (type (simple-array function ,*vo-option-num*) voice-current-value-function-vec))
      (setf voice-current-value-function-vec
            (nth-current-value-function-vec current-value-function-vec voice-number))
      (symbol-macrolet 
           ,(loop with key
                  for i below *vo-option-num*
                  do (setf key (aref *vo-option-key-vec* i))
                  collect `(,(symbol-of-key key)
                            ,(if (member key (coerce *vo-non-dynamic-option-key-vec* 'list))
                               `(aref voice-option-val-vec ,i)
                               `(funcall (aref voice-current-value-function-vec ,i)
                                         (aref voice-option-val-vec ,i)))))
           ,. vo-processform))))



;;; Main single voice process 


(defun voice-output (rhy-pat pitch-pat inst vo-input-data voice-number)
   (with-dynamic-vo-symbol-reference
      (process with dur = 0 
           with wait-dur = 0
           ; absolute times must be multiplied for midi-files with spec. tempo
           with mf-factor = (/ mf-tempo 60) 
           with by-mf-factor = (/ 1 mf-factor)
           with actual-lb
           with actual-ub
           with index-list 
           with error-cases 
           with last-key-list 
           with last-index-lists 
           with buffer-length = 0
           with filled-buffer?  ; indicates when buffer-length has reached max-buffer-length at start
           with by-level-change-step-time = (/ 1 (* level-change-step-time mf-factor))
           with int-level = (integerize-amp level)
           with corrected-level-change-delay
           with corrected-level-change-break
           with last-level = int-level
           with pitch-condition
           with overrule-step-pat
           with amp = (if (numberp dyn) dyn)
           with last-amp
           with item-base
           with corrected-dur
           with interpl-data-list
           with step-time-num-list
           initially
           (progn 
             ; closure including conditions for pitch-pattern filter
             (setf pitch-condition 
               #'(lambda (key-list)
                   (block pc
                    (if (eod? key-list) (return-from pc key-list))
                    (setf key-list (strong-listify key-list)
                          actual-lb (or lb (car (nth inst (inst-key-bounds-list 
                                                           *current-inst-midi-setup*))))
                          actual-ub (or ub (cadr (nth inst (inst-key-bounds-list 
                                                           *current-inst-midi-setup*)))))
                    (if runtime-data-check
                        (mapc #'(lambda (x) (runtime-item-check x :pitch voice-number 0)) 
                              key-list))
                    
                    (if transpose-to-bounds 
                        (setf key-list 
                              (mapcar #'(lambda (x) 
                                          (transpose-key-to-bounds x (list actual-lb actual-ub))) 
                                             key-list)))
                    (setf index-list (mapcar #'key->index key-list))
                    (cond (overrule-step-pat 
                           (return-from pc (unlistify key-list)))
                          ((and actual-lb (some #'(lambda (x) (< x actual-lb)) key-list))
                           (push 1 error-cases))
                          ((and actual-ub (some #'(lambda (x) (> x actual-ub)) key-list))
                           (push 2 error-cases))
                          ((and pitch-rep-horizon
                                (/= pitch-rep-horizon 0)    
                                last-index-lists
                                (member index-list 
                                        (subseq last-index-lists 0 (if filled-buffer? 
                                                                       pitch-rep-horizon
                                                                       buffer-length))
                                        :test #'intersection))
                           (push (if (cadr index-list) 4 3) error-cases))
                          ((and octave-step-horizon
                                (/= octave-step-horizon 0)    
                                last-index-lists
                                (member index-list 
                                        (subseq last-index-lists 0 (if filled-buffer? 
                                                                       octave-step-horizon
                                                                       buffer-length))
                                        :test #'octave-steps?))
                           (push (if (cadr index-list) 6 5) error-cases))
                          ((and avoid-primes-group
                                (cadr index-list)
                                (pairs-in-list? index-list #'=))
                           (push 7 error-cases))
                          ((and avoid-octaves-group
                                (cadr index-list)
                                (octaves-in-index-list? index-list))
                           (push 8 error-cases))
                          ((and avoid-primes-group
                                (some #'(lambda (x) 
                                          (primes-in-group? x avoid-primes-group)) 
                                      index-list))
                           (push 9 error-cases))
                          ((and avoid-octaves-group
                                (some #'(lambda (x) 
                                          (real-octaves-in-group? x avoid-octaves-group)) 
                                      index-list))
                           (push 10 error-cases))
                          ((and max-step
                                last-index-lists
                                (> (abs (list-step-width (car last-index-lists) index-list))
                                   max-step))
                           (push 11 error-cases))
                          (t (return-from pc (unlistify key-list))))
                   (incf (aref *conflict-count* voice-number (car error-cases)))
                   nil)))

             ; setting filter predicate for pitch pattern

             (global-restriction-on-filter pitch-pat
                   (if step-start      ; case stepping pattern
                       (let ((last-pitch (pattern-last-value pitch-pat)) 
                             last-scale-indices current-start-pitch (first-pitch-calc t)
                             (round-delta (round-delta *current-inst-midi-setup*))) 
                         ; predicate for absolute pitches as result of stepping:
                          #'(lambda (x) 
                              (if runtime-data-check
                                  (mapc #'(lambda (y) (runtime-item-check y :step voice-number 0)) 
                                        (strong-listify x)))
                              (let* (possible-pitch result last-start-pitch pat-type shadow-pat
                                                    last-start-status new-step-start)
                                (multiple-value-setq 
                                  (last-start-pitch last-start-status pat-type shadow-pat) step-start)
                                (setf new-step-start  
                                      (cond ((null pat-type)
                                             (or first-pitch-calc))
                                            (t (or last-start-status))))
                                (if new-step-start
                                    (setf current-start-pitch 
                                          last-start-pitch))
                                
                                (setf overrule-step-pat
                                      (and new-step-start strict-step-start)
                                      possible-pitch 
                                      (let ((steps (let ((next-steps (listify x)))
                                                     (if overrule-step-pat
                                                       (mapcar #'(lambda (y) (declare (ignore y)) 0)
                                                               next-steps)
                                                       next-steps)))
                                            (pitches (listify 
                                                      (if new-step-start
                                                          current-start-pitch
                                                          last-pitch))))
                                        (multiple-value-bind (scale new-scale?) pitch-scale
                                          (if (null scale)
                                              (mapcar #'+ pitches steps)
                                              (progn
                                                (cond ((or overrule-step-pat new-scale? new-step-start)
                                                       (setf last-scale-indices
                                                             (mapcar #'(lambda (y z) 
                                                                         (next-position-in-new-scale 
                                                                          y z scale round-delta))
                                                                     pitches steps)))
                                                      (t (setf last-scale-indices
                                                               (mapcar #'(lambda (y z) 
                                                                           (next-position-in-old-scale 
                                                                            y z scale))
                                                                       last-scale-indices steps)))) 
                                                (if overrule-step-pat
                                                    pitches
                                                    (mapcar #'(lambda (y) (aref scale y)) 
                                                            last-scale-indices)))))))        
                                (setf result 
                                      (funcall pitch-condition possible-pitch))
                                (when result 
                                      (setf last-pitch result)
                                      (if (eql pat-type 1)
                                          (setf (slot-value shadow-pat 'ref-pat-call) nil))
                                      (if first-pitch-calc
                                          (setf first-pitch-calc nil))
                                      result))))
                       pitch-condition) ; <- case ordinary pitch-pattern
                   max-next
                   #'(lambda () 
                       (setf *error-cases-history* error-cases
                             *error-voice-number* voice-number
                             *error-max-next* max-next) 
                       (unsolvable-condition-message t)
                       (error "")
                       )))

           ; start of actual output process

           while (and (< (now) (* max-time mf-factor)) 
                      (progn
                        (if runtime-data-check
                            (next-with-check rhy-pat :rhythm voice-number 0)
                            (next rhy-pat))
                        (when (not (arrayp (pattern-last-value rhy-pat)))
                              ; pitch pat only nexted if no rest in rhythm-pat
                              ; then nexting happens under chunk of conditions (pitch-condition) 
                              (setf last-amp amp)
                              (mapc (if runtime-data-check 
                                        #'(lambda (x) (next-with-check (cdr x) (car x) voice-number 0))
                                        #'(lambda (x) (next (cdr x)))) 
                                    voice-event-option-key+pat-alist)
                              (setf amp dyn)
                              (next pitch-pat))
                        (not (or (eod? rhy-pat)
                                 (eod? dyn)
                                 (eod? pitch-pat)))))
               do
               (setf wait-dur (* (time-parse (time-value (pattern-last-value rhy-pat)) 
                                             tempo beat abs-times) mf-factor))
               (when (not (arrayp (pattern-last-value rhy-pat)))
                     (setf dur (* wait-dur dur-factor)
                           error-cases nil
                           last-key-list (listify (pattern-last-value pitch-pat))
                           last-index-lists (lifo-buffer index-list last-index-lists max-buffer-length))
                     (when (not filled-buffer?)
                           (if (< buffer-length max-buffer-length) (incf buffer-length))
                           (if (= buffer-length max-buffer-length) (setf filled-buffer? t)))
                     
                     (if (arrayp amp)
                       (progn          ; prepare data for level-change-messages
                           (multiple-value-setq (corrected-level-change-delay corrected-dur
                                                 corrected-level-change-break)
                             (prepare-duration dur 
                                               (if (> (aref amp 0) last-level)              
                                                     (* level-change-delay mf-factor)  
                                                     0)
                                               (* level-change-break mf-factor) 
                                               level-shape-shrink))
                           (multiple-value-setq (interpl-data-list step-time-num-list item-base)
                             (make-dynamic-interpl-data corrected-dur amp by-level-change-step-time 
                                                        min-level-change-steps))
                           (setf last-level (car (last interpl-data-list)))
                           (sprout (inst-amp-change inst interpl-data-list step-time-num-list
                                       (or item-base base)) 
                                   (+ (now) corrected-level-change-delay)))
                       (when (arrayp last-amp)  ; level normalization if cr/dim before 
                             (setf corrected-level-change-delay 
                                   (if (<= amp last-level)              
                                       0
                                       (prepare-duration dur (* level-change-delay mf-factor) 
                                         (* level-change-break mf-factor) 
                                         level-shape-shrink)))
                             (setf last-level int-level)
                             (sprout (set-inst-levels (listify inst) int-level) 
                                         (+ (now) corrected-level-change-delay))))
                           
                     (loop for index in index-list
                           for key in (listify (pattern-last-value pitch-pat)) 
                           for k from 0 do
                           
                           (sprout (lock-index index voice-number) (now))                
                           (sprout (unlock-index index voice-number) 
                                   (+ (now) dur (- (* unlock-backward-shift mf-factor))))
                           (sprout (lock-pc-index index voice-number) (now))                
                           (sprout (unlock-pc-index index voice-number) 
                                   (+ (now) dur (- (* unlock-backward-shift mf-factor))))
                           (output (make-instance 'setup-tied-midi 
                                                         :keynum key :index index
                                                         :duration dur :time (now)
                                                         :by-mf-factor by-mf-factor
                                                         :amplitude (amp-value amp int-level) 
                                                         :amp-array (if (arrayp amp) amp)
                                                         :inst inst
                                                         :setup *current-inst-midi-setup*  
                                                         :print-midi print
                                                         :voice voice-number
                                                         :output-type output-type))))
               (wait wait-dur))))



;;; Virtual output (arbitrary data, but maybe midi, midimsgs or list of them, too)


(defclass v-event (event)
  ((data :initform nil 
         :initarg data :initarg :data)
   (duration :initform nil 
             :initarg duration :initarg :duration)
   (by-mf-factor :initform nil 
                 :initarg by-mf-factor :initarg :by-mf-factor)
   (v-voice :initform nil 
            :initarg v-voice :initarg :v-voice)
   (print-data :initform nil 
          :initarg print-event :initarg :print-data)))


(defmethod write-event ((object v-event) io time)
  (if (slot-value object 'print-data) 
      (format (slot-value object 'print-data) 
              "~%      virtual-voice-nr~3,,,'_@A:  time~11,5,,,'_F  dur~11,5,,,'_F   data__________~A" 
              (slot-value object 'v-voice)
              (* (slot-value object 'by-mf-factor) (slot-value object 'time))
              (* (slot-value object 'by-mf-factor) (slot-value object 'duration))
              (slot-value object 'data)
              object))
  (if (every #'(lambda (x) (member (type-of x) '(midimsg midi)))
             (listify (slot-value object 'data)))
      (loop for event in (listify (slot-value object 'data)) do
            (write-event event io time))))



(defprocess v-output (rhy-pat data-pat option-keys mf-factor &key v-tempo v-beat v-abs-times
                               v-print-stream v-max-time v-voice-number)
  (process with dur = 0
           with mf-v-max-time = (* v-max-time mf-factor)
           with by-mf-factor = (/ 1 mf-factor)
           while (and (< (now) mf-v-max-time)
                      (progn
                        (next rhy-pat)
                        (if (not (arrayp (pattern-last-value rhy-pat)))
                            (if option-keys
                                (next-with-check data-pat option-keys v-voice-number 1)
                                (next data-pat)))
                        (not (or (eod? rhy-pat)
                                 (eod? data-pat)))))
               do
               (setf dur (* (time-parse (time-value (pattern-last-value rhy-pat)) 
                                                   v-tempo v-beat v-abs-times) mf-factor))
               (when (not (arrayp (pattern-last-value rhy-pat)))
                     (output (make-instance 'v-event
                               :data (pattern-last-value data-pat)
                               :duration dur
                               :by-mf-factor by-mf-factor
                               :time (now)
                               :print-data v-print-stream
                               :v-voice v-voice-number)))  
               (wait dur)))


;;; Mixing and time-shifting processes 


(defun pat-events-calc (pitch-pats rhythm-pats instruments pe-options pe-args)
  (with-all-slots pe-options pat-events-options 
    (let* ((density (length pitch-pats))
           (v-density (if (listp v-data-pats) (length v-data-pats) 1))
           (voice-queue (or voice-order (range-list (1- density))))
           (vq (if voice-order voice-queue))  
           (v-voice-queue (or v-voice-order (range-list (1- v-density))))
           (v-vq (if v-voice-order v-voice-queue))
           (mf-factor (/ mf-tempo 60))
           (shift (mapcar #'(lambda (x) (* x voice-shift mf-factor)) 
                         (range-list (+ v-density density) 1)))
           vo-input-data
           print-streams 
           v-print-streams
           v-data-option-keys
           all-print-streams
           voice-output-closures)
      (sleep sleep)
      (setf *conflict-count* (make-array (list density 15) :initial-element 0))
      (clear-indicator-vecs)
      (unwind-protect 
        (progn
          (multiple-value-setq (print-streams v-print-streams all-print-streams)
            (combine-make-open-print-streams print v-print))
          (setf print print-streams
                vo-input-data (make-vo-input-data pe-options pe-args density)
                v-data-option-keys (collect-v-data-option-keys vo-input-data v-data-pats))
          ; (describe vo-input-data))
          (setf voice-output-closures 
                (loop for i below density collect
                      (voice-output (if-nth i rhythm-pats vq) (if-nth i pitch-pats vq) 
                                    (if-nth i instruments vq)  
                                    vo-input-data (nth i voice-queue)))) 
          
          (events 
           (mix-processes   ; mix of voices and virtual-voices 
            (nconc              
             (if (not (equalp to "midi.port"))
                 (list (send-inst-midi-setup-data (listify instruments) *current-inst-midi-setup*)))
             (list (set-inst-levels (listify instruments) 
                                    (unlistify (mapcar #'integerize-amp (listify level)))))
             (mapcar #'(lambda (i)
                         (v-output (if-nth i v-rhythm-pats v-vq) (if-nth i v-data-pats v-vq) 
                                   (if-nth i v-data-option-keys v-vq) mf-factor
                                   :v-tempo (if-nth i v-tempo v-vq) :v-max-time (if-nth i v-max-time v-vq)
                                   :v-beat (if-nth i v-beat v-vq) :v-abs-times (if-nth i v-abs-times v-vq)
                                   :v-print-stream (if-nth i v-print-streams v-vq) 
                                   :v-voice-number (nth i v-voice-queue)))
                     (range-list (1- v-density)))
             voice-output-closures)
            (if (not (equalp to "midi.port"))
                (list* 0 0 shift)
                (list* 0 shift)))
           (if (not (equalp to "midi.port"))
               (insure-io to :tempo mf-tempo :timesig mf-measure)
               to)) 
          (loop for stream in (cons t all-print-streams) do
                (conflict-protocol voice-queue stream))
          (format t "Output was sent to ~S ~2%" to))
        (when *error-cases-history*
              (loop for stream in all-print-streams do
                    (unsolvable-condition-message stream))
              (setf *error-cases-history* nil
                    *error-voice-number* nil
                    *error-max-next* nil))
        (close-print-streams all-print-streams)))))



(defun pat-events (rhythm-pats pitch-pats instruments &rest args &key dyn max-next 
                              no-8 no-1 voice-order tempo beat abs-times print 
                              transpose-to-bounds voice-shift max-time max-buffer-length 
                              unlock-backward-shift dur-factor no-reps no-8-steps to 
                              step-start strict-step-start max-step lb ub
                              pitch-scale sleep level 
                              level-change-step-time min-level-change-steps base 
                              level-shape-shrink level-change-delay level-change-break
                              v-voice-order v-data-pats v-rhythm-pats v-print 
                              v-tempo v-beat v-abs-times v-max-time 
                              (input-data-check t) runtime-data-check mf-tempo mf-measure)
  (declare (ignore dyn no-8 no-1 voice-order tempo beat abs-times print 
                   transpose-to-bounds voice-shift max-time max-buffer-length 
                   unlock-backward-shift dur-factor no-reps no-8-steps to 
                   strict-step-start step-start max-step lb ub pitch-scale sleep level  
                   level-change-step-time min-level-change-steps base 
                   level-shape-shrink level-change-delay level-change-break
                   v-data-pats v-rhythm-pats v-print v-tempo v-beat v-abs-times
                   v-max-time v-voice-order runtime-data-check mf-tempo mf-measure))
  (if (not (boundp '*current-inst-midi-setup*))
      (error "no current setup - use 'install-setup'"))
  (if input-data-check 
      (apply #'pat-events-input-check pitch-pats rhythm-pats instruments args))
    
  ; modifying input

  (let* ((pe-options (make-pe-options-object args))
         (filter-pitch-pats (listify pitch-pats))
         (density (length filter-pitch-pats)))
    (loop for i below density do  ; pitch pattern as filter pattern 
            (setf (nth i filter-pitch-pats) 
                  (new filter :of (nth i filter-pitch-pats)
                       :name 
                       (concatenate 'string "pitch-" 
                                    (write-to-string i))
                       :limit (if-nth i max-next))))
    (if (not (member :to args :test #'symbol-name-eql?))  
        (setf (slot-value pe-options 'to)
              (if (direct-midi-out *current-inst-midi-setup*)
                  "midi.port"
                  (error "midifile-namestring required"))))
    
    (when (not (equalp (slot-value pe-options 'to) "midi.port")) 
          (if (not (member :mf-tempo args :test #'symbol-name-eql?))
              (setf (slot-value pe-options 'mf-tempo)
                    (slot-value pe-options 'tempo))))

          
    (loop for (x y) in '((:tempo :v-tempo)
                         (:beat :v-beat)
                         (:max-time :v-max-time)
                         (:abs-times :v-abs-times)
                         (:print :v-print)) do
          (subordinate-pe-options y x args pe-options))
    ;(describe pe-options)
    (redefine-step-start pe-options "stst")
    (redefine-pe-grouping pe-options 'no-1 density)
    (redefine-pe-grouping pe-options 'no-8 density)
    (loop for option in '(no-reps no-8-steps) do
          (setf (slot-value pe-options option)
                (map-anyway (slot-value pe-options option) 
                            #'(lambda (x &optional y) (declare (ignore y)) 
                               (if (pattern? x) x (buffer-horizon x))))))
    (redefine-pe-buffer-length pe-options density)
    (funcall #'pat-events-calc filter-pitch-pats (listify rhythm-pats) 
             instruments pe-options args)))



